@@ -2,16 +2,20 @@
 
 """Core functionality
 
-Author: G.J.J. van den Burg
 License: See LICENSE file.
-Copyright: 2020, G.J.J. van den Burg
 
 """
 
+import logging
 import os
-import warnings
 import sqlite3
 import shutil
+import datetime as dt
+
+from .dbproto import StructuredReaction
+from .dbproto import StructuredReactions
+
+from .addressbook import make_addressbook
 
 from .exceptions import DatabaseNotFound
 from .exceptions import DatabaseVersionNotFound
@@ -19,17 +23,20 @@ from .exceptions import DatabaseVersionNotFound
 from .models import Attachment
 from .models import MMSMessageRecord
 from .models import Quote
+from .models import Reaction
 from .models import Recipient
-from .models import RecipientId
 from .models import SMSMessageRecord
 from .models import Thread
 
 from .html import dump_thread
-from .html_colors import get_random_color
+
+from .versioninfo import VersionInfo
+
+logger = logging.getLogger(__name__)
 
 
-def check_backup(backup_dir):
-    """Check that we have the necessary files"""
+def check_backup(backup_dir) -> VersionInfo:
+    """Check that we have the necessary files and return VersionInfo"""
     if not os.path.join(backup_dir, "database.sqlite"):
         raise DatabaseNotFound
     if not os.path.join(backup_dir, "DatabaseVersion.sbf"):
@@ -37,14 +44,12 @@ def check_backup(backup_dir):
     with open(os.path.join(backup_dir, "DatabaseVersion.sbf"), "r") as fp:
         version_str = fp.read()
 
-    # We have only ever seen database version 23, so we don't proceed if it's
-    # not that. Testing and pull requests welcome.
     version = version_str.split(":")[-1].strip()
-    if not version in ["23", "65", "80", "89"]:
-        warnings.warn(
-            f"Warning: Found untested Signal database version: {version}."
-        )
-    return version
+    versioninfo = VersionInfo(version)
+
+    if not versioninfo.is_tested_version():
+        logger.warn(f"Found untested Signal database version: {version}.")
+    return versioninfo
 
 
 def get_color(db, recipient_id):
@@ -57,86 +62,7 @@ def get_color(db, recipient_id):
     return color
 
 
-def make_recipient_v23(db, recipient_id):
-    """ Create a Recipient instance from a given recipient id (db version 23)"""
-    if recipient_id.startswith("__textsecure_group__"):
-        qry = db.execute(
-            "SELECT title FROM groups WHERE group_id=?", (recipient_id,)
-        )
-        label = qry.fetchone()
-        isgroup = True
-    else:
-        qry = db.execute(
-            "SELECT system_display_name "
-            "FROM recipient_preferences "
-            "WHERE recipient_ids=?",
-            (recipient_id,),
-        )
-        label = qry.fetchone()
-        isgroup = False
-
-    name = str(recipient_id) if label[0] is None else label[0]
-    color = get_color(db, recipient_id)
-
-    phone = str(recipient_id)
-    rid = RecipientId(recipient_id)
-    return Recipient(rid, name=name, color=color, isgroup=isgroup, phone=phone)
-
-
-def make_recipient_v80(db, recipient_id):
-    """Create a Recipient instance from a recipient id (db version 65, 80, 89)"""
-    qry = db.execute(
-        "SELECT group_id, "
-        "phone, "
-        "system_display_name, "
-        "profile_joined_name, "
-        "color "
-        "FROM recipient "
-        "WHERE _id=?",
-        (recipient_id,),
-    )
-    groupid, phone, name, joined_name, color = qry.fetchone()
-    if color is None:
-        color = get_random_color()
-
-    isgroup = groupid is not None
-    if isgroup:
-        qry = db.execute(
-            "SELECT title FROM groups WHERE group_id=?", (groupid,)
-        )
-        res = qry.fetchone()
-        if res is not None:
-            name = res[0]
-        else:
-            warnings.warn(
-                f"Group for recipient {recipient_id} is {groupid} which does not exist."
-            )
-
-    if name is None:
-        if joined_name:
-            name = joined_name
-        elif phone:
-            name = phone
-        else:
-            name = ""
-
-    rid = RecipientId(recipient_id)
-    return Recipient(rid, name=name, color=color, isgroup=isgroup, phone=phone)
-
-
-def make_recipient(db, recipient_id, version=None):
-    if version == "23":
-        return make_recipient_v23(db, recipient_id)
-    elif version in ["65", "80", "89"]:
-        return make_recipient_v80(db, recipient_id)
-
-    warnings.warn(
-        f"Untested database version {version}, defaulting to latest known working version."
-    )
-    return make_recipient_v80(db, recipient_id)
-
-
-def get_sms_records(db, thread, version=None):
+def get_sms_records(db, thread, addressbook):
     """ Collect all the SMS records for a given thread """
     sms_records = []
     sms_qry = db.execute(
@@ -146,9 +72,10 @@ def get_sms_records(db, thread, version=None):
     )
     qry_res = sms_qry.fetchall()
     for _id, address, date, date_sent, body, _type in qry_res:
+        sms_auth = addressbook.get_recipient_by_address(address)
         sms = SMSMessageRecord(
             _id=_id,
-            addressRecipient=make_recipient(db, address, version=version),
+            addressRecipient=sms_auth,
             recipient=thread.recipient,
             dateSent=date_sent,
             dateReceived=date,
@@ -165,8 +92,8 @@ def get_attachment_filename(_id, unique_id, backup_dir, thread_dir):
     fname = f"Attachment_{_id}_{unique_id}.bin"
     source = os.path.abspath(os.path.join(backup_dir, fname))
     if not os.path.exists(source):
-        warnings.warn(
-            f"Warning: couldn't find attachment {source}. Maybe it was deleted?"
+        logger.warn(
+            f"Couldn't find attachment '{source}'. Maybe it was deleted or never downloaded."
         )
         return None
 
@@ -201,14 +128,53 @@ def add_mms_attachments(db, mms, backup_dir, thread_dir):
         mms.attachments.append(a)
 
 
+def get_mms_reactions(encoded_reactions, addressbook, mid):
+    reactions = []
+    if encoded_reactions:
+        try:
+            structured_reactions = StructuredReactions.loads(encoded_reactions)
+        except (ValueError, IndexError, TypeError) as e:
+            logger.warn(
+                f"Failed to load reactions for message {mid}: {str(e)}"
+            )
+            return []
+
+        for structured_reaction in structured_reactions.reactions:
+            recipient = addressbook.get_recipient_by_address(
+                str(structured_reaction.who)
+            )
+            reaction = Reaction(
+                recipient=recipient,
+                what=structured_reaction.what,
+                time_sent=dt.datetime.fromtimestamp(
+                    structured_reaction.time_sent // 1000
+                ),
+                time_received=dt.datetime.fromtimestamp(
+                    structured_reaction.time_received // 1000
+                ),
+            )
+            reaction.time_sent = reaction.time_sent.replace(
+                microsecond=(structured_reaction.time_sent % 1000) * 1000
+            )
+            reaction.time_received = reaction.time_received.replace(
+                microsecond=(structured_reaction.time_received % 1000) * 1000
+            )
+            reactions.append(reaction)
+
+    return reactions
+
+
 def get_mms_records(
-    db, thread, recipients, backup_dir, thread_dir, version=None
+    db, thread, addressbook, backup_dir, thread_dir, versioninfo
 ):
     """ Collect all MMS records for a given thread """
     mms_records = []
+
+    reaction_expr = versioninfo.get_reactions_query_column()
+
     qry = db.execute(
         "SELECT _id, address, date, date_received, body, quote_id, "
-        "quote_author, quote_body, msg_box FROM mms WHERE thread_id=?",
+        f"quote_author, quote_body, msg_box, {reaction_expr} FROM mms WHERE thread_id=?",
         (thread._id,),
     )
     qry_res = qry.fetchall()
@@ -222,30 +188,16 @@ def get_mms_records(
         quote_author,
         quote_body,
         msg_box,
+        reactions,
     ) in qry_res:
-        quote = None
-        if quote_id:
-            quote_auth = next(
-                (
-                    r
-                    for r in recipients
-                    if str(r.recipientId._id) == str(quote_author)
-                ),
-                None,
-            )
-            if quote_auth is None:
-                # Quote is from someone who isn't a recipient (e.g. a friend
-                # quotes a third person in a group). We'll just create a
-                # recipient object for this person.
-                rid = RecipientId(quote_author)
-                quote_auth = Recipient(
-                    rid, quote_author, color=None, isgroup=False
-                )
-            quote = Quote(_id=quote_id, author=quote_auth, text=quote_body)
+        quote = get_mms_quote(addressbook, quote_id, quote_author, quote_body)
 
+        decoded_reactions = get_mms_reactions(reactions, addressbook, _id)
+
+        mms_auth = addressbook.get_recipient_by_address(address)
         mms = MMSMessageRecord(
             _id=_id,
-            addressRecipient=make_recipient(db, address, version=version),
+            addressRecipient=mms_auth,
             recipient=thread.recipient,
             dateSent=date,
             dateReceived=date_received,
@@ -253,6 +205,7 @@ def get_mms_records(
             body=body,
             quote=quote,
             attachments=[],
+            reactions=decoded_reactions,
             _type=msg_box,
         )
         mms_records.append(mms)
@@ -263,13 +216,26 @@ def get_mms_records(
     return mms_records
 
 
+def get_mms_quote(addressbook, quote_id, quote_author, quote_body):
+    quote = None
+    if quote_id:
+        quote_auth = addressbook.get_recipient_by_address(quote_author)
+        quote = Quote(_id=quote_id, author=quote_auth, text=quote_body)
+    return quote
+
+
 def populate_thread(
-    db, thread, recipients, backup_dir, thread_dir, version=None
+    db, thread, addressbook, backup_dir, thread_dir, versioninfo=None
 ):
     """ Populate a thread with all corresponding messages """
-    sms_records = get_sms_records(db, thread, version=version)
+    sms_records = get_sms_records(db, thread, addressbook)
     mms_records = get_mms_records(
-        db, thread, recipients, backup_dir, thread_dir, version=version
+        db,
+        thread,
+        addressbook,
+        backup_dir,
+        thread_dir,
+        versioninfo,
     )
     thread.sms = sms_records
     thread.mms = mms_records
@@ -279,27 +245,28 @@ def process_backup(backup_dir, output_dir):
     """ Main functionality to convert database into HTML """
 
     # Verify backup and open database
-    db_version = check_backup(backup_dir)
+    versioninfo = check_backup(backup_dir)
     db_file = os.path.join(backup_dir, "database.sqlite")
     db_conn = sqlite3.connect(db_file)
     db = db_conn.cursor()
 
+    # Get and index all contact and group names
+    addressbook = make_addressbook(db, versioninfo)
+
     # Start by getting the Threads from the database
     query = db.execute("SELECT _id, recipient_ids FROM thread")
     threads = query.fetchall()
 
-    # Now turn the recipients from the threads into Recipient classes
-    recipients = []
-    for _, recipient_id in threads:
-        r = make_recipient(db, recipient_id, version=db_version)
-        recipients.append(r)
-
     # Combine the recipient objects and the thread info into Thread objects
-    for (_id, _), recipient in zip(threads, recipients):
+    for (_id, recipient_id) in threads:
+        recipient = addressbook.get_recipient_by_address(recipient_id)
+        if recipient is None:
+            logger.warn(f"No recipient with address {recipient_id}")
+
         t = Thread(_id=_id, recipient=recipient)
         thread_dir = os.path.join(output_dir, t.sanename)
         populate_thread(
-            db, t, recipients, backup_dir, thread_dir, version=db_version
+            db, t, addressbook, backup_dir, thread_dir, versioninfo=versioninfo
         )
         dump_thread(t, thread_dir)
 
