@@ -2,9 +2,7 @@
 
 """Code for writing out the HTML
 
-Author: G.J.J. van den Burg
 License: See LICENSE file.
-Copyright: 2020, G.J.J. van den Burg
 
 """
 
@@ -32,14 +30,29 @@ def is_all_emoji(body):
     return len(emoji_list(body)) == len(body) and len(body) > 0
 
 
-def format_emoji(body, is_quote=False):
-    """ Wrap emoji in <span> so we can style it easily """
+def format_message(body, is_quote=False):
+    """Format message by processing all characters.
+
+    - Wrap emoji in <span> for styling them
+    - Escape special HTML chars
+    """
     emoji_pos = emoji_list(body)
     new_body = ""
     emoji_lookup = {p["location"]: p["emoji"] for p in emoji_pos}
+    skip = 0
     for i, c in enumerate(body):
-        if i in emoji_lookup:
+        if skip > 0:
+            # Skip additional characters from multi-character emoji
+            skip = skip - 1
+        elif i in emoji_lookup:
             new_body += "<span class='msg-emoji'>%s</span>" % emoji_lookup[i]
+            skip = len(emoji_lookup[i]) - 1
+        elif c == "&":
+            new_body += "&amp;"
+        elif c == "<":
+            new_body += "&lt;"
+        elif c == ">":
+            new_body += "&gt;"
         else:
             new_body += c
     return new_body
@@ -63,7 +76,7 @@ def dump_thread(thread, output_dir):
 
     # Create the message color CSS (depends on individuals)
     group_color_css = ""
-    msg_css = ".msg-sender-%i { background: %s; }\n"
+    msg_css = ".msg-sender-%i { /* recipient id: %5s */ background: %s;}\n"
     if is_group:
         group_recipients = set(m.addressRecipient for m in messages)
         sender_idx = {r: k for k, r in enumerate(group_recipients)}
@@ -81,15 +94,25 @@ def dump_thread(thread, output_dir):
                     None,
                 )
                 ar_color = ar.color if color is None else color
-            group_color_css += msg_css % (idx, COLORMAP[ar_color])
+            group_color_css += msg_css % (
+                idx,
+                ar.rid,
+                COLORMAP[ar_color],
+            )
             colors_used.append(ar.color)
     else:
+        # Retrieve sender info from an incoming message, if any
         firstInbox = next(
             (m for m in messages if is_inbox_type(m._type)), None
         )
-        clr = firstInbox.addressRecipient.color if firstInbox else "teal"
-        clr = "teal" if clr is None else clr
-        group_color_css += msg_css % (0, COLORMAP[clr])
+        if firstInbox:
+            clr = firstInbox.addressRecipient.color
+            clr = "teal" if clr is None else clr
+            group_color_css += msg_css % (
+                0,
+                firstInbox.addressRecipient.rid,
+                COLORMAP[clr],
+            )
 
     # Create a simplified dict for each message
     prev_date = None
@@ -125,7 +148,7 @@ def dump_thread(thread, output_dir):
         # Deal with quoted messages
         quote = {}
         if isinstance(msg, MMSMessageRecord) and msg.quote:
-            quote_author_id = msg.quote.author.recipientId._id
+            quote_author_id = msg.quote.author.rid
             quote_author_name = msg.quote.author.name
             if quote_author_id == quote_author_name:
                 name = "You"
@@ -133,7 +156,7 @@ def dump_thread(thread, output_dir):
                 name = quote_author_name
             quote = {
                 "name": name,
-                "body": format_emoji(msg.quote.text),
+                "body": format_message(msg.quote.text),
                 "attachments": [],
             }
 
@@ -143,7 +166,7 @@ def dump_thread(thread, output_dir):
             all_emoji = not msg.quote and is_all_emoji(body)
         else:
             all_emoji = is_all_emoji(body)
-        body = format_emoji(body)
+        body = format_message(body)
 
         # Create message dictionary
         aR = msg.addressRecipient
@@ -159,9 +182,10 @@ def dump_thread(thread, output_dir):
             "name": aR.name,
             "sender_idx": sender_idx[aR] if is_group else "0",
             "quote": quote,
+            "reactions": [],
         }
 
-        # Add attachments
+        # Add attachments and reactions
         if isinstance(msg, MMSMessageRecord):
             for a in msg.attachments:
                 if a.quote:
@@ -169,6 +193,17 @@ def dump_thread(thread, output_dir):
                 else:
                     out["attachments"].append(a)
 
+            for r in msg.reactions:
+                out["reactions"].append(
+                    {
+                        "recipient_id": r.recipient.rid,
+                        "name": r.recipient.name,
+                        "what": r.what,
+                        "time_sent": r.time_sent,
+                        "time_received": r.time_received,
+                    }
+                )
+
         simple_messages.append(out)
 
     if not simple_messages:
